// Flocking
// Daniel Shiffman <http://www.shiffman.net>
// The Nature of Code, Spring 2009

// Boid class
// Methods for Separation, Cohesion, Alignment sumed
class Flocking {

  float SEPWEIGHT = 8.0f;
  float SEPDIST = 30f;
  float ALIWEIGHT = 80.0f;
  float ALIDIST = 50f;
  float COHWEIGHT = 1.0f;
  float COHDIST = 100f;
  float MAXSPEED = 50.0f;
  float MAXFORCE = 0.5f;
	Vector3D loc;
	Vector3D vel;
	Vector3D acc;
	float r;
	float sphericalPosTheta = 0;
	float sphericalPosPhi = 0;
	double sphereRadius = 200;
	float zoneVal = 0;
	String debug = "";

	Flocking(Vector3D _loc) {
		acc = new Vector3D(0, 0, 0);
		vel = new Vector3D((float) Math.random() * 2 - 1, (float) Math.random() * 2 - 1, (float) Math.random() * 2 - 1);
		loc = _loc;
		r = 2.0f;
	}

	void move(Flocking[] others) {
		flock(others);
		update();
		borders();
	}

	// We accumulate a new acceleration each time based on three rules
	void flock(Flocking[] others) {
		Vector3D sep = separate(others);   // Separation
		Vector3D ali = align(others);      // Alignment
		Vector3D coh = cohesion(others);   // Cohesion
		// Arbitrarily weight these forces
		sep.mult(SEPWEIGHT);
		ali.mult(ALIWEIGHT);
		coh.mult(COHWEIGHT);
		// sum the force vectors to acceleration
		acc.sum(sep);
		acc.sum(ali);
		acc.sum(coh);
	}

	// Method to update location
	void update() {
		// Update velocity
		vel.sum(acc);
		// Limit speed
		vel.limit(MAXSPEED);
		loc.sum(vel);
		// Reset acceleration to 0 each cycle
		acc.mult(0);
	}

	void seek(Vector3D target) {
		acc.sum(steer(target, false));
	}

	void arrive(Vector3D target) {
		acc.sum(steer(target, true));
	}

	// A method that calculates a steering vector towards a target
	// Takes a second argument, if true, it slows down as it approaches the target
	Vector3D steer(Vector3D target, boolean slowdown) {
		Vector3D steer;  // The steering vector
		Vector3D desired = Vector3D.diff(target, loc);  // A vector pointing from the location to the target
		float d = desired.length(); // Distance from the target is the lengthnitude of the vector
		// If the distance is greater than 0, calc steering (otherwise return zero vector)
		if (d > 0) {
			// Normalize desired
			desired.normalize();
			// Two options for desired vector lengthnitude (1 -- based on distance, 2 -- MAXSPEED)
			if ((slowdown) && (d < 100.0f)) {
				desired.mult(MAXSPEED * (d / 100.0f)); // This damping is somewhat arbitrary
			} else {
				desired.mult(MAXSPEED);
			}
			// Steering = Desired minus Velocity
			steer = Vector3D.diff(desired, vel);
			steer.limit(MAXFORCE);  // Limit to maximum steering force
		} else {
			steer = new Vector3D(0, 0, 0);
		}
		return steer;
	}

	// Wraparound
	void borders() {
		int depth = Config.width / 4;
		if (loc.x() < -r) {
			loc.setX(Config.width + r);
		}
		if (loc.y() < -r) {
			loc.setY(Config.height + r);
		}
		if (loc.z() < 0) {
			loc.setZ(depth);
		}
		if (loc.x() > Config.width + r) {
			loc.setX(-r);
		}
		if (loc.y() > Config.height + r) {
			loc.setY(-r);
		}
		if (loc.z() > depth) {
			loc.setZ(0);
		}
	}

	// Separation
	// Method checks for nearby others and steers away
	Vector3D separate(Flocking[] others) {
		Vector3D steer = new Vector3D(0, 0, 0);
		int count = 0;
		// For every boid in the system, check if it's too close
		for (int i = 0; i < others.length; i++) {
			Flocking other = others[i];
			if (other.loc.x() > loc.x() + SEPDIST) {
				break;
			}
			// If the distance is greater than 0 and less than an arbitrary amount (0 when you are yourself)
			if ((other.loc.x() == loc.x() && other.loc.y() == loc.y())
							|| other.loc.x() < loc.x() - SEPDIST
							|| other.loc.y() < loc.y() - SEPDIST
							|| other.loc.y() > loc.y() + SEPDIST) {
				continue;
			}
			// Calculate vector pointing away from neighbor
			Vector3D diff = Vector3D.diff(loc, other.loc);
			float d = Vector3D.dist(loc, other.loc);
			diff.normalize();
			diff.div(d);        // Weight by distance
			steer.sum(diff);
			count++;            // Keep track of how many
		}
		// Average -- divide by how many
		if (count > 0) {
			steer.div((float) count);
		}
		// As long as the vector is greater than 0
		if (steer.length() > 0) {
			// Implement Reynolds: Steering = Desired - Velocity
			steer.normalize();
			steer.mult(MAXSPEED);
			steer.diff(vel);
			steer.limit(MAXFORCE);
		}
		return steer;
	}

	// Alignment
	// For every nearby boid in the system, calculate the average velocity
	Vector3D align(Flocking[] others) {
		Vector3D steer = new Vector3D(0, 0, 0);
		int count = 0;
		for (int i = 0; i < others.length; i++) {
			Flocking other = others[i];
			if (other.loc.x() > loc.x() + ALIDIST) {
				break;
			}
			if ((other.loc.x() == loc.x() && other.loc.y() == loc.y())
							|| other.loc.x() < loc.x() - ALIDIST
							|| other.loc.y() < loc.y() - ALIDIST
							|| other.loc.y() > loc.y() + ALIDIST) {
				continue;
			}
			steer.sum(other.vel);
			count++;
		}
		if (count > 0) {
			steer.div((float) count);
		}
		// As long as the vector is greater than 0
		if (steer.length() > 0) {
			// Implement Reynolds: Steering = Desired - Velocity
			//steer.normalize();
			steer.mult(MAXSPEED);
			steer.diff(vel);
			steer.limit(MAXFORCE);
		}
		return steer;
	}

	// Cohesion
	// For the average location (i.e. center) of all nearby others, calculate steering vector towards that location
	Vector3D cohesion(Flocking[] others) {
		Vector3D sum = new Vector3D(0, 0, 0);   // Start with empty vector to accumulate all locations
		int count = 0;
		for (int i = 0; i < others.length; i++) {
			Flocking other = others[i];
			if (other.loc.x() > loc.x() + COHDIST) {
				break;
			}
			if ((other.loc.x() == loc.x() && other.loc.y() == loc.y())
							|| other.loc.x() < loc.x() - COHDIST
							|| other.loc.y() < loc.y() - COHDIST
							|| other.loc.y() > loc.y() + COHDIST) {
				continue;
			}
			sum.sum(other.loc); // sum location
			count++;
		}
		if (count > 0) {
			sum.div((float) count);
			return steer(sum, true);  // Steer towards the location
		}
		return sum;
	}

	public void setZoneVal(float _zoneVal) {
		zoneVal = _zoneVal;
	}

	public float getX() {
		return loc.x();
	}

	public float getY() {
		return loc.y();
	}

	public float getThetaRad() {
		return vel.heading2D();
	}

	public float getThetaDeg() {
		return 180 * vel.heading2D() / (float) (Math.PI);
	}

	public void computeSphericalCoords() {
		sphericalPosTheta = (getX() * 360f / Config.width) * (float) (Math.PI / 180);
		sphericalPosPhi = (getY() * 180f / Config.height) * (float) (Math.PI / 180);
	}

	public Vector3D getCartesianFromSphericalCoords() {
		computeSphericalCoords();
//    float sphericalX = (float) ((sphereRadius+zoneVal*200) * SinCosLUT.sin(sphericalPosTheta) * SinCosLUT.cos(sphericalPosPhi));
//    float sphericalY = (float) ((sphereRadius+zoneVal*200) * SinCosLUT.sin(sphericalPosTheta) * SinCosLUT.sin(sphericalPosPhi));
//    float sphericalZ = (float) ((sphereRadius+zoneVal*200) * SinCosLUT.cos(sphericalPosTheta));
		float sphericalX = (float) (sphereRadius * SinCosLUT.sin(sphericalPosTheta) * SinCosLUT.cos(sphericalPosPhi));
		float sphericalY = (float) (sphereRadius * SinCosLUT.sin(sphericalPosTheta) * SinCosLUT.sin(sphericalPosPhi));
		float sphericalZ = (float) (sphereRadius * SinCosLUT.cos(sphericalPosTheta));
		return new Vector3D(sphericalX, sphericalY, sphericalZ);
	}
}
