
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.LinkedList;

public class Graph {

  public ArrayList<Node> nodeArray = new ArrayList<Node>();
  public ArrayList<Node> selectedNodes = new ArrayList<Node>();
  public ArrayList<NodeSet> nodeSets = new ArrayList<NodeSet>();
  private boolean isForceDirectedLayoutActive = true;
  int SPRING_REST_LENGTH = 50;

  public Graph() {
  }

  public int getNumNodes() {
    return nodeArray.size();
  }

  public Node getNode(int index) {
    return nodeArray.get(index);
  }

  public ArrayList<Node> getNodes() {
    return nodeArray;
  }

  public void addNode(Node n) {
    if (!nodeArray.contains(n)) {
      nodeArray.add(n);
    }
  }

  public void addEdge(Node n1, Node n2) {
    if (n1 == n2 || n1 == null || n2 == null) {
      return;
    }
    if (!n1.neighbours.contains(n2)) {
      n1.neighbours.add(n2);
    }
    if (!n2.neighbours.contains(n1)) {
      n2.neighbours.add(n1);
    }
  }

  public void addEdge(int nodeIndex1, int nodeIndex2) {
    addEdge(getNode(nodeIndex1), getNode(nodeIndex2));
  }

  // Note that this won't remove the node from the set of selected nodes;
  // that must be done by the caller, if so desired.
  private void deleteNode(Node n) {
    for (int i = 0; i < n.neighbours.size(); ++i) {
      Node otherNode = n.neighbours.get(i);
      otherNode.neighbours.remove(n);
    }
    n.neighbours.clear();
    nodeArray.remove(n);
    //TODO ... remove from any NodeSets the node belongs to
  }

  public void removeEdge(Node n1, Node n2) {
    if (n1 == n2 || n1 == null || n2 == null) {
      return;
    }
    if (n1.neighbours.contains(n2)) {
      n1.neighbours.remove(n2);
    }
    if (n2.neighbours.contains(n1)) {
      n2.neighbours.remove(n1);
    }
  }

  public int getNumSelectedNodes() {
    return selectedNodes.size();
  }

  public boolean isNodeSelected(Node node) {
    return selectedNodes.contains(node);
  }

  public ArrayList<Node> getSelectedNodes() {
    return selectedNodes;
  }

  public void selectNode(Node node) {
    if (!isNodeSelected(node)) {
      selectedNodes.add(node);
    }
  }

  public void deselectNode(Node node) {
    selectedNodes.remove(node);
  }

  public void selectByLabel(String s) {
    for (int j = 0; j < nodeArray.size(); ++j) {
      Node n = nodeArray.get(j);
      if (n.label.equals(s)) {
        selectNode(n);
      }
    }
  }

 
  public void selectAllNodes() {
    selectedNodes.clear();
    for (int i = 0; i < nodeArray.size(); ++i) {
      selectedNodes.add(nodeArray.get(i));
    }
  }

  public void deselectAllNodes() {
    selectedNodes.clear();
  }

  public void invertSelectedNodes() {
    ArrayList<Node> newlySelectedNodes = new ArrayList<Node>();
    for (int i = 0; i < nodeArray.size(); ++i) {
      Node n = nodeArray.get(i);
      if (!selectedNodes.contains(n)) {
        newlySelectedNodes.add(n);
      }
    }
    selectedNodes = newlySelectedNodes;
  }

  public void deleteSelectedNodes() {
    for (int i = 0; i < selectedNodes.size(); ++i) {
      deleteNode(selectedNodes.get(i));
    }
    selectedNodes.clear();
  }

  // The return value of this can be used to index into nodeArray.
  public int getIndexOfNode(Node n) {
    assert n != null;

    // If the cached index is NOT correct ...
    int index = n.getIndex();
    if (index < 0 || index >= nodeArray.size() || getNode(index) != n) {
      // ... recompute all the indices
      // (this takes linear time, but subsequent calls
      // for this and other nodes will be fast,
      // until the next time the cached indices are
      // made invalid because some nodes have been
      // added or deleted)
      for (int i = 0; i < nodeArray.size(); ++i) {
        nodeArray.get(i).setIndex(i);
      }
    }

    // assert that the cached index IS now correct
    assert getNode(n.getIndex()) == n;

    // return the cached index
    return n.getIndex();
  }

  public boolean areNodesAdjacent(Node n1, Node n2) {
    if (n1.neighbours.size() > n2.neighbours.size()) {
      return n2.neighbours.contains(n1);
    }
    return n1.neighbours.contains(n2);
  }

  public Node findNearestNode(float x, float y, float minDistanceThreshold) {
    Node nearestNode = null;
    float smallestDistanceSquared = 0;
    for (int i = 0; i < nodeArray.size(); ++i) {
      Node n = nodeArray.get(i);
      float dx = x - n.x;
      float dy = y - n.y;
      float distanceSquared = dx * dx + dy * dy;
      if (nearestNode == null || distanceSquared < smallestDistanceSquared) {
        smallestDistanceSquared = distanceSquared;
        nearestNode = n;
      }
    }
    if (smallestDistanceSquared <= minDistanceThreshold * minDistanceThreshold) {
      return nearestNode;
    }
    return null;
  }

  public void setShapeTypeOfNodes(ArrayList<Node> nodes, int nodeShape) {
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      n.nodeShape = nodeShape;
    }
  }

  public void setShowLabelsOfNodes(ArrayList<Node> nodes, boolean flag) {
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      n.showLabel = flag;
    }
  }

  public void toggleShowLabelsOfNodes(ArrayList<Node> nodes) {
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      n.showLabel = !n.showLabel;
    }
  }

  public void setFixedFlagOfNodes(ArrayList<Node> nodes, boolean flag) {
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      n.isFixed = flag;
    }
  }

  public void randomizePositionsOfNodes(
          // If null, the randomization is applied to all nodes in the network,
          // otherwise it is only applied to the given ``nodes''.
          ArrayList<Node> nodes,
          // If true, even fixed nodes are randomized.
          boolean randomizeFixedNodes,
          // If true, the random displacements are proportional
          // to the extents of the current layout of the nodes.
          // If false, the randomization is chosen to uniformly
          // fill the given ``boundingRectangle''.
          boolean localized,
          // If non-null, the randomized positions will not fall
          // outside this rectangle.
          Rect2D boundingRectangle) {
    if (nodes == null) {
      nodes = nodeArray;
    }
    Rect2D rectangle = new Rect2D();
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      if (randomizeFixedNodes || !n.isFixed) {
        rectangle.bound(new Vector2D(n.x, n.y));
      }
    }
    float distance = rectangle.getDiagonal().length();
    if (distance < SPRING_REST_LENGTH / 2) {
      distance = SPRING_REST_LENGTH / 2;
    }
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      if (randomizeFixedNodes || !n.isFixed) {
        if (localized || boundingRectangle == null) {
          n.x += distance * (Math.random() - 0.5f);
          n.y += distance * (Math.random() - 0.5f);
        } else {
          n.x = (float) (boundingRectangle.getMin().x() + Math.random() * boundingRectangle.getDiagonal().x());
          n.y = (float) (boundingRectangle.getMin().y() + Math.random() * boundingRectangle.getDiagonal().y());
        }

        if (boundingRectangle != null) {
          // bound the node's position

          if (n.x < boundingRectangle.getMin().x()) {
            n.x = boundingRectangle.getMin().x();
          } else if (n.x > boundingRectangle.getMax().x()) {
            n.x = boundingRectangle.getMax().x();
          }

          if (n.y < boundingRectangle.getMin().y()) {
            n.y = boundingRectangle.getMin().y();
          } else if (n.y > boundingRectangle.getMax().y()) {
            n.y = boundingRectangle.getMax().y();
          }
        }
      }
    }
  }

  /**
   * Returns the bounding rectangle of the nodes in 2D. If the given vector
   * of nodes is null, calculates and returns the bounding rectangle of the
   * entire network.
   */
  public Rect2D getBoundingRectangle(ArrayList<Node> nodes) {
    if (nodes == null) {
      nodes = nodeArray;
    }
    Rect2D rectangle = new Rect2D();
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      rectangle.bound(new Vector2D(n.x, n.y));
    }
    return rectangle;
  }

   public Vector2D computeGeometricCentreOfNodes(
          // If null, the centre of all nodes in the network is returned,
          // otherwise only the given ``nodes'' are considered.
          ArrayList<Node> nodes) {
    return getBoundingRectangle(nodes).getCentre();
  }

  /**
   * Computers the centroid of the given nodes in 2 dimensions, or the entire network if
   * the given vector is null.
   */
  public Vector2D computeCentroidOfNodes(
          // If null, the centroid of all nodes in the network is returned,
          // otherwise only the given ``nodes'' are considered.
          ArrayList<Node> nodes) {
    if (nodes == null) {
      nodes = nodeArray;
    }
    Vector2D centroid = new Vector2D(0, 0);
    int numNodes = nodes.size();
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      centroid.v[0] += n.x;
      centroid.v[1] += n.y;
    }
    centroid = Vector2D.mult(centroid, 1.0f / numNodes);
    return centroid;
  }

  /**
   * Centres the nodes in two dimenstion about the given centre. If the given set
   * of vectors is null, centres all the nodes in the network about this point.
   */
  public void centreNodes(ArrayList<Node> nodes, Vector2D newCentre) {
    if (nodes == null) {
      nodes = nodeArray;
    }
    Vector2D currentCentre = computeGeometricCentreOfNodes(nodes);
    Vector2D shift = Vector2D.diff(newCentre, currentCentre);
    for (int i = 0; i < nodes.size(); ++i) {
      Node n = nodes.get(i);
      n.x += shift.x();
      n.y += shift.y();
    }
  }

  private void clearAllNodeSets() {
    for (int i = 0; i < nodeSets.size(); ++i) {
      nodeSets.get(i).clear();
    }
    nodeSets.clear();
  }
}


